/* TEF: tiny electric fence on Cygwin:
  Copyright (c) 2006   Katsuhiko Gondow (gondow(at)cs.titech.ac.jp)
  $Id: tef.c,v 1.1 2006/06/07 09:29:20 gondow Exp $ */

/* NOTE: using the following GCC option inteposes the calling
   of __wrap_malloc between the callee and caller of malloc.

   % gcc -Wl,-wrap=malloc -Wl,-wrap=realloc -Wl,-wrap=free  ...

   Unfortunately, cygwin1.dll uses the symbol 'malloc' as 
   a strong symbol (not weak one), which results in segmentation
   fault in cygwin1.dll's malloc_init().
   To work around this problem, I introduced a variable 
   'TEF_is_in_main'. TEF works iff TEF_is_in_main is 1, so
   please add the following line in the beginning of the main.

      TEF_is_in_main = 1;

   Second problem is that some system functions in Cygwin like
   asprintf and strdup use malloc_r, not malloc. Malloc_r is
   not exported, so there is no way to hook it. I made a code
   of asprintf and strdup.

   Finally, the array 'tef[NTEF]' is statically allocated
   (to simply avoid using malloc), so many calls of malloc
   can break down TEF's execution. Of course, you can rewrite
   this issue by using mmap().
*/

#if defined(USE_TEF)
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <assert.h>
#include <sys/mman.h>
#include <sys/types.h>

#include "tef.h"

// #define TEF_CHECK_UNDERFLOW

#define DEBUG_BREAK asm ("int3");
#if !defined(MIN)
#define MIN(a,b)   (((a)<(b))?(a):(b))
#endif

struct TEF {
    char *page_top, *page_protected;
    char *returned_ptr;
    unsigned int magic, size;
    struct TEF *next, *prev;
};

enum {
    ALIGNMENT    = 8,
    PAGESIZE     = 1024 * 64,
    NTEF         = 1024 * 20,
    MAGIC_USED   = 0xDEADBEAFU,
    MAGIC_UNUSED = 0xCAFEBABEU,
};

int TEF_is_in_main = 0;

/* ROUNDUP assumes that n is a power of 2. */
#define ROUNDUP(size, align)    (((size)+((align)-1)) & ~((align)-1))
#define ALLOC_SIZE(size)        (ROUNDUP(size, PAGESIZE) + (PAGESIZE)) 

static void TEF_dump          (struct TEF *tef) UNUSED;
static void TEF_init          (void);
static void TEF_insert_head   (struct TEF *head, struct TEF *tef);
static void TEF_insert_tail   (struct TEF *head, struct TEF *tef);
static void TEF_insert_sorted (struct TEF *tef);
static int  TEF_is_null       (void);
static void TEF_remove        (struct TEF *tef);

static void *TEF_alloc_for_overflow  (unsigned int size);
static void *TEF_alloc_for_underflow (unsigned int size);
static void *TEF_realloc             (void *ptr, unsigned int size);
static void TEF_dealloc              (void *ptr);


static struct TEF tef_head;
static struct TEF tef [NTEF];
static int TEF_is_initialized = 0;

/* ================================================================== */
static void
TEF_dump  (struct TEF *tef)
{
    fprintf (stderr, "%p, %p, %p, %u, %x, %p, %p, %p\n",
	     tef->page_top, tef->returned_ptr, tef->page_protected,
	     tef->size, tef->magic, tef, tef->prev, tef->next);
}


static struct TEF *
get_free_tef (void)
{
    int i;
    for (i = 0; i < NTEF; i++) {
	if (tef [i].magic == MAGIC_UNUSED) {
	    tef [i].magic = MAGIC_USED;
	    return &tef [i];
	}
    }

    fprintf (stderr, "TEF ERROR: tef pool is exhaused\n");
    DEBUG_BREAK
    return NULL;
}

static void *
TEF_alloc_for_overflow (unsigned int size)
{
    int ret;
    struct TEF *tef = get_free_tef ();
    assert (tef->magic == MAGIC_USED);
    tef->size     = size;
    tef->page_top = mmap (0, ALLOC_SIZE(size), PROT_WRITE | PROT_READ,
			  MAP_ANONYMOUS | MAP_PRIVATE, 0, 0);
    if (tef->page_top == MAP_FAILED) {
	perror ("mmap"); exit (1);
    }
    tef->page_protected = tef->page_top + PAGESIZE;
    ret = mprotect (tef->page_protected, PAGESIZE, PROT_NONE);
    if (ret == -1) {
	perror ("mprotect"); exit (1);
    }

    tef->next = tef->prev = tef;
    tef->returned_ptr = tef->page_protected - ROUNDUP(size, ALIGNMENT);
    TEF_insert_sorted (tef);

    return tef->returned_ptr;
}

static void *
TEF_alloc_for_underflow (unsigned int size)
{
    int ret;
    struct TEF *tef = get_free_tef ();
    assert (tef->magic == MAGIC_USED);
    tef->size     = size;
    tef->page_top = mmap (0, ALLOC_SIZE(size), PROT_WRITE | PROT_READ,
			  MAP_ANONYMOUS | MAP_PRIVATE, 0, 0);
    if (tef->page_top == MAP_FAILED) {
	perror ("mmap"); exit (1);
    }
    tef->page_protected = tef->page_top;
    ret = mprotect (tef->page_protected, PAGESIZE, PROT_NONE);
    if (ret == -1) {
	perror ("mprotect"); exit (1);
    }

    tef->next = tef->prev = tef;
    tef->returned_ptr = tef->page_top + PAGESIZE;
    TEF_insert_sorted (tef);

    return tef->returned_ptr;
}

static void *
TEF_realloc (void *ptr, unsigned int size)
{
    struct TEF *p;
    void *ptr_new;
    
    if (ptr == NULL)
	return __wrap_malloc (size);

    for (p = tef_head.next; p != &tef_head; p = p->next) {
	if ((char *)ptr < p->returned_ptr) {
	    break;
	}
	if ((char *)ptr == p->returned_ptr) { /* found */
	    ptr_new = __wrap_malloc (size);
	    memmove (ptr_new, ptr, MIN(p->size, size));
	    __wrap_free (ptr);
	    return ptr_new;
	}
    }

    fprintf (stderr, "TEF ERROR: not allocated before: %p\n", ptr);
    DEBUG_BREAK
    exit (1);
    return NULL;
}

static void
TEF_dealloc (void *ptr)
{
    struct TEF *p;

    for (p = tef_head.next; p != &tef_head; p = p->next) {
	if ((char *)ptr < p->returned_ptr) {
	    break;
	}
	if ((char *)ptr == p->returned_ptr) { /* found */
	    int ret = munmap (p->page_top, ALLOC_SIZE(p->size));
	    if (ret == -1) {
		perror ("munmap"); exit (1);
	    }

	    TEF_remove (p);

	    p->page_top = p->page_protected = p->returned_ptr = NULL;
	    assert (p->magic == MAGIC_USED);
	    p->magic = MAGIC_UNUSED;
	    p->size = 0;
	    p->next = p->prev = NULL;
	    return;
	}
    }

    fprintf (stderr, "TEF ERROR: illegal free: %p\n", ptr);
    asm ("int3");
    __real_free (ptr);
}

static void
TEF_init   (void)
{
    int i;
    tef_head.next = tef_head.prev = &tef_head;
    for (i = 0; i < NTEF; i++) {
	tef [i].magic = MAGIC_UNUSED;
    }

    TEF_is_initialized = 1;
}

static void
TEF_insert_head (struct TEF *head, struct TEF *tef)
{
    tef->prev = head;
    tef->next = head->next;
    head->next->prev = tef;
    head->next = tef;
}

static void
TEF_insert_tail (struct TEF *head, struct TEF *tef)
{
    tef->prev = head->prev;
    tef->next = head;
    head->prev->next = tef;
    head->prev = tef;
}

static void
TEF_insert_sorted (struct TEF *tef)
{
    struct TEF *p;

    if (TEF_is_null ()) {
	TEF_insert_head (&tef_head, tef);
	return;
    }

    for (p = tef_head.next; p != &tef_head; p = p->next) {
	if ((p->returned_ptr == tef->returned_ptr)
	    || (p->page_top == tef->page_top)) {
	    fprintf (stderr, "TEF ERROR: this chunk already inserted\n");
	    DEBUG_BREAK
	    exit (1);
	}
	if (p->returned_ptr > tef->returned_ptr) {
	    TEF_insert_head (p->prev, tef);
	    return;
	}
    }

    TEF_insert_tail (&tef_head, tef);
}

static void
TEF_remove  (struct TEF *tef)
{
    tef->next->prev = tef->prev;
    tef->prev->next = tef->next;
    tef->next = tef->prev = NULL;
}

static int
TEF_is_null  (void)
{
    return &tef_head == tef_head.next;
}
/* ================================================================== */
void
TEF_dump_list (void) 
{
    int i;
    struct TEF *p;

    fprintf (stderr, "head: ");
    TEF_dump (&tef_head);

    for (i = 0, p = tef_head.next; p != &tef_head; i++, p = p->next) {
	fprintf (stderr, "%d: ", i);
	TEF_dump (p);
    }
}

void
TEF_dump_array (void) 
{
    int i;

    fprintf (stderr, "head: ");
    TEF_dump (&tef_head);

    for (i = 0; i < NTEF; i++) {
	fprintf (stderr, "%d: ", i);
	TEF_dump (&tef [i]);
    }
}

void *
__wrap_malloc (size_t size)
{ 
#if 0
    fprintf (stderr, "I am __wrap_malloc: %d\n", TEF_is_in_main);
#endif

    if (!TEF_is_in_main) {
	return __real_malloc (size);
    }

    if (TEF_is_initialized == 0) {
	TEF_init ();
    }


#if defined(TEF_CHECK_UNDERFLOW)
    return TEF_alloc_for_underflow (size);
#else
    return TEF_alloc_for_overflow (size);
#endif
}

void *
__wrap_realloc (void *ptr, size_t size)
{
#if 0
    fprintf (stderr, "I am __wrap_realloc: %d\n", TEF_is_in_main);
#endif

    if (!TEF_is_in_main) {
	return __real_realloc (ptr, size);
    }

    if (TEF_is_initialized == 0) {
	TEF_init ();
    }
    return TEF_realloc (ptr, size);
}

void
__wrap_free (void *ptr)
{
#if 0
    fprintf (stderr, "I am __wrap_free: %d\n", TEF_is_in_main);
#endif

    if (!TEF_is_in_main) {
	__real_free (ptr);
	return;
    }

    if (TEF_is_initialized == 0) {
	fprintf (stderr, "TEF ERROR: free before alloc: %p\n", ptr);
	__real_free (ptr);
	return;
    }
    TEF_dealloc (ptr);
}

#if 0
void
test1 (void)
{
    int *p;
    p = __wrap_malloc (sizeof (int));
    *p = 123;
    printf ("%d\n", p[0]);
    //printf ("%d\n", p[2]);
    __wrap_free (p);
}

void
test2 (void)
{
    int *p;
    p = __wrap_malloc (sizeof (int));
    *p = 123;
    printf ("%d\n", p[0]);
    printf ("%d\n", p[-1]);
    __wrap_free (p);
}

void 
test3 (void)
{
    char *p;
    asprintf (&p, "*** %s ***\n", "fufufu");
    free (p);
}

void 
test4 (void)
{
    char *p = "hogehoge", *q;
    q = strdup (p);
    free (q);
}

int
main (void)
{
    TEF_is_in_main = 1;
    test4 ();
    return 0;
}
#endif

#endif /* USE_TEF */
