/* tfix: fixing ID values
   in TBCppA project: tracer-based CPP analyzer.
   Copyright (c) 2006   Katsuhiko Gondow (gondow(at)cs.titech.ac.jp)
   $Id: tfix.c,v 1.4 2006/06/08 08:07:26 gondow Exp $ */

#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <unistd.h>
#include <libxml/xmlreader.h>
#include <libxml/xmlwriter.h>

#include "misc.h"
/* ==================================================================== */
struct fentry {
    char   *file_id;
    struct fentry *next, *prev;
};

static struct fentry fentry_head = {NULL, &fentry_head, &fentry_head};
static xmlTextReaderPtr reader;
static xmlTextWriterPtr writer;
/* ==================================================================== */
static void fentry_dump  (void) __attribute__((unused));
static void fentry_insert_head (struct fentry *p);
static void check_id (char *name, char *id);
static void tfix (const char *outfile);
/* ==================================================================== */
static void
fentry_dump  (void)
{
    struct fentry *p;
    fprintf (stderr, "===================================\n");
    for (p = fentry_head.next; p != &fentry_head; p = p->next) {
	fprintf (stderr, "id=%s\n", p->file_id);
    }
}

static void
fentry_insert_head (struct fentry *p)
{
    p->prev = &fentry_head;
    p->next = fentry_head.next;
    fentry_head.next->prev = p;
    fentry_head.next = p;
}

static void
check_id (char *name, char *id)
{
    struct fentry *p;
    for (p = fentry_head.next; p != &fentry_head; p = p->next) {
	if (!strcmp (id, p->file_id)) {
	    return;
	}
    }
    p = emalloc (sizeof (struct fentry));
    p->file_id = estrdup (id);
    p->next = p->prev = p;
    fentry_insert_head (p);
    xmlTextWriterWriteAttribute (writer, name, id);
}

static void
tfix (const char *outfile)
{
    reader = xmlReaderForFd    (STDIN_FILENO, NULL, NULL, 0);
    xmlTextReaderSetParserProp (reader, XML_PARSER_SUBST_ENTITIES, 0);
    xmlTextReaderSetParserProp (reader, XML_PARSER_VALIDATE, 0);
    xmlTextReaderSetParserProp (reader, XML_PARSER_LOADDTD, 0);
    xmlTextReaderSetParserProp (reader, XML_PARSER_DEFAULTATTRS, 0);
    

    if (reader == NULL) {
	fprintf (stderr, "cannot open: stdin\n");
	exit (1);
    }

    /* ???: how can I open a writer via STDOUT_FILENO */
    writer = xmlNewTextWriterFilename (outfile, 0);
    if (writer == NULL) {
	fprintf (stderr, "cannot open: %s\n", outfile);
	exit (1);
    }

    xmlTextWriterWriteString (writer, "<?xml version=\"1.0\"?>\n");

    while(xmlTextReaderRead (reader)) {
	char *name = xmlTextReaderName (reader);
	char *val  = xmlTextReaderValue (reader);
	int  type;

	switch (type = xmlTextReaderNodeType (reader)) {
	case XML_READER_TYPE_ELEMENT:
	    xmlTextWriterStartElement (writer, name);

	    if (xmlTextReaderHasAttributes (reader)) {
		while (xmlTextReaderMoveToNextAttribute (reader)) {
		    char *name = xmlTextReaderName (reader);
		    char *val  = xmlTextReaderValue (reader);
		    if (!strcmp (name, "id")) {
			check_id (name, val);
		    } else {
			xmlTextWriterWriteAttribute (writer, name, val);
		    }
		    xmlFree (name);
		    xmlFree (val);
		}
		xmlTextReaderMoveToElement (reader);
	    }
	    
	    if (xmlTextReaderIsEmptyElement (reader)) {
		xmlTextWriterEndElement (writer);
	    }
	    break;

	case XML_READER_TYPE_END_ELEMENT:
	    xmlTextWriterEndElement (writer);
	    break;

	case XML_READER_TYPE_TEXT:
	    xmlTextWriterWriteString (writer, val);
	    break;

	case XML_READER_TYPE_DOCUMENT_TYPE:
#if 0
	    xmlTextWriterWriteFormatDTD (writer, name, NULL, "TBCppA.dtd",
					 NULL);
	    xmlTextWriterWriteFormatRaw (writer, "\n");
#else
	    {
		char buf [1024];
		xmlDocPtr doc = xmlTextReaderCurrentDoc (reader);
		xmlDtdPtr dtd = xmlGetIntSubset (doc);
		/* .doctype will be used in 'tstrip.xsl' */
		FILE *fp = fopen (".doctype", "wb+");

		xmlElemDump (fp, doc, (xmlNodePtr)dtd);
		fflush (fp);
		rewind (fp);
		while (fgets (buf, sizeof (buf), fp) != NULL) {
		    xmlTextWriterWriteFormatRaw (writer, "%s", buf);
		}
		fclose (fp);
		xmlTextWriterWriteFormatRaw (writer, "\n");
	    }
#endif
	    break;
	case XML_READER_TYPE_SIGNIFICANT_WHITESPACE:
	    xmlTextWriterWriteFormatRaw (writer, "%s", val);
	    break;

	case XML_READER_TYPE_COMMENT:
	    xmlTextWriterWriteComment (writer, val);
	    break;

	case XML_READER_TYPE_ENTITY_REFERENCE:
	    xmlTextWriterWriteFormatRaw (writer, "&%s;", name);
	    break;

	case XML_READER_TYPE_ENTITY:
	case XML_READER_TYPE_NONE:
	case XML_READER_TYPE_ATTRIBUTE:
	case XML_READER_TYPE_XML_DECLARATION:
	case XML_READER_TYPE_CDATA:
	case XML_READER_TYPE_PROCESSING_INSTRUCTION:
	case XML_READER_TYPE_DOCUMENT:
	case XML_READER_TYPE_DOCUMENT_FRAGMENT:
	case XML_READER_TYPE_NOTATION:
	case XML_READER_TYPE_WHITESPACE:
	case XML_READER_TYPE_END_ENTITY:
	    fprintf (stderr, "tfix: unsupported type: %d\n", type);
	    assert (0);
	    break;
	default:
	    fprintf (stderr, "tfix: unknown type: %d\n", type);
	    assert (0);
	    break;
	}

	xmlFree (name);
	xmlFree (val);
    }

    xmlTextWriterWriteFormatRaw (writer, "\n");
    xmlFreeTextReader (reader);
    xmlFreeTextWriter (writer);
}

/* Don't forget to do 'xmlFree" for a pointer of the type "xmlChar". */
/* ==================================================================== */
int
main (int argc, char *argv [])
{
    if (argc != 2) {
	fprintf (stderr, "Usage: %s out.xml < in.xml\n", argv [0]);
	exit (1);
    }

    tfix (argv [1]);
    return 0;
}

